#
#    Copyright (c) 2009 Tom Keffer <tkeffer@gmail.com>
#
#    See the file LICENSE.txt for your full rights.
#
import math
import time
import datetime
import os.path
import syslog
import Cheetah.Template
import weewx
import weewx.stats
import weeutil.Moon
import weeutil.Sun
import weeutil.Almanac


def htmlGen(config_dict, launchtime_ts = None, timestamp = None) :
    """Generate HTML pages for a given timestamp.
    
    This routine is completely stateless outside of what's held in
    the SQL databases: everything is recalculated
    each time. Perhaps that makes it inefficient, but it allows it
    to be easily multithreaded.
    
    config_dict: An instance of ConfigObj.
    
    launchtime_ts: A timestamp of when weewx was launched. Used to
    calculate uptime. [optional. Default is to not calculate uptime]
    
    timestamp: The time for which the HTML pages are to be generated.
    [optional. Default is to generate pages for the last good timestamp
    in the database archive.]
    
    """
    ngen = 0
    t1 = time.time()

    # Statistics (hi, low, avg, etc.) will be held in this dictionary,
    # keyed by the time period ('current', 'day', 'month', 'year', and 'rainyear') 
    stats = {}

    # Open up the main database archive
    archiveFilename = os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                   config_dict['Archive']['archive_file'])
    archive = weewx.archive.Archive(archiveFilename)

    if timestamp is None:
        timestamp = archive.lastGoodStamp()
        
    # Get the 'current' conditions from it:
    stats['current'] = archive.getRecord(timestamp)

    # Open up the stats database
    statsFilename = os.path.join(config_dict['Station']['WEEWX_ROOT'], 
                                 config_dict['Stats']['stats_file'])
    statsDb = weewx.stats.StatsDb(statsFilename)
        
    # Get 'day', 'month', 'year', and 'rainyear' statistics from it:
    rain_year_start = int(config_dict['Station'].get('rain_year_start', '1'))
    stats['day'], stats['week'], stats['month'], stats['year'], stats['rainyear'] = statsDb.getAllStats(timestamp, rain_year_start)
    
    # How to format types:
    formatter = Formatter(config_dict['Labels']['ImperialFormats'],
                          config_dict['Labels']['ImperialUnits'],
                          config_dict['Labels']['Time'])
    
    # Create a view into the stats using a formatter:
    statsView = ModelView(stats, formatter)

    # (Mostly) static data about the station. Seldom changes.
    station = Station(config_dict, launchtime_ts)
    
    # Calculate an appropriate Almanac for this station:
    almanac = weeutil.Almanac.Almanac(timestamp, 
                                      station.latitude_f, station.longitude_f,
                                      config_dict['Almanac'].get('moon_phases', None))

    # Get the directory holding the templates
    template_dir = os.path.join(config_dict['Station']['WEEWX_ROOT'],
                                config_dict['HTML']['template_root'])
    # Get the directories that will hold the generated HTML files:
    html_dir     = os.path.join(config_dict['Station']['WEEWX_ROOT'],
                                config_dict['HTML']['html_root'])

    for timescale in ('index', 'week', 'month', 'year'):

        template_file = os.path.join(template_dir, timescale + ".tmpl")

        html = Cheetah.Template.Template(file = template_file, 
                                         searchList = [{'station' : station, 'almanac' : almanac}, statsView])

        file = open(os.path.join(html_dir, timescale + ".html"), mode='w')
        print >>file, html
        ngen += 1

    t2 = time.time()
    elapsed_time = t2-t1
    syslog.syslog(syslog.LOG_INFO, "HtmlGenerator: generated %d HTML pages in %.3f seconds" % (ngen, elapsed_time))
 

class Station(object):
    """Static data about the station. Rarely changes.
    
    """
    def __init__(self, config_dict, launchtime_ts = None):
        """Extracts info from the config_dict and stores it in self.
        
        """
        # TODO: Should allow I18N of 'N', 'S', etc. labels
        self.latitude_f = config_dict['Station'].as_float('latitude')
        self.latitude = latlon_string(self.latitude_f, 'N' if self.latitude_f>0 else 'S')
        self.longitude_f = config_dict['Station'].as_float('longitude')
        self.longitude = latlon_string(self.longitude_f, 'E' if self.longitude_f>0 else 'W')
        self.altitude_f = config_dict['Station'].as_float('altitude')
        # TODO: Shouldn't hard wire in the reference to 'ImperialUnits':
        alt_units = config_dict['Labels']['ImperialUnits']['altitude']
        self.altitude = "%d %s" % (self.altitude_f, alt_units)
        self.location = config_dict['Station']['location']
        self.rain_year_start = config_dict['Station']['rain_year_start']
        self.radar_url = config_dict['Station'].get('radar_url','')
        self.uptime = weeutil.weeutil.secs_to_string(time.time() - launchtime_ts)
        self.version = weewx.__version__
        # The following works on Linux only:
        try:
            os_uptime_secs =  float(open("/proc/uptime").read().split()[0])
            self.os_uptime = weeutil.weeutil.secs_to_string(int(os_uptime_secs + 0.5))
        except:
            self.os_uptime = ''
    
def latlon_string(ll, hemi):
    labs = abs(ll)
    (frac, deg) = math.modf(labs)
    min = frac * 60.0
    str = "%d\xb0%0.2f'%s" % (deg, min, hemi)
    return str.decode('string_escape')
    
    
class ModelView(object):
    """A view into an hierarchical data model.
    
    Much of the HTML rendering logic depends on this class. Given a 
    hierarchical data model consisting of dictionaries and attributes,
    this class offers views into it, using introspection. The views
    are through a formatter object. The views are created on the fly, 
    on demand. Thus the model can change without changing any state in the 
    view. It does this by creating itself recursively, as it goes
    down the model, root to leaf node. When it finally hits an integer
    or float, it stops and invokes the formatter.
    
    For example, given a model such as:
        data = {}
        data['year'] = {}
        data['year']['outTemp'] = {}
        data['year']['outTemp']['min'] = 45.88
        data['year']['outTemp']['max'] = 55.298
    
    Then a ModelView can be constructed into it, using a formatter:
    
        formatter = weewx.html.Formatter({'outTemp' : '%.1f'}, 
                                         {'outTemp' : " Fahrenheit"}, None)
        dataView = ModelView(data, formatter)
    
        print data['year']['outTemp']['min']     # Prints 45.88
        print dataView['year']['outTemp']['min'] # Prints "45.9 Fahrenheit"

    Note that data['year']['outTemp']['min'] returns a *float* of value 45.88,
    (which print then converts into a string), while the view offers a formatted view, 
    returning a *string*.
    
    """
    def __init__(self, model, formatter, context = ()):
        """Initialize an instance of ModelView.
        
        model: The hierarchical model to be viewed.
        
        formatter: An instance of weewx.html.Formatter, or something that looks like it.
        
        context: Contains the accumulating context as ModelView creates 
        itself recursively. 
        """
        self.model     = model
        self.formatter = formatter
        self.context   = context
        
    def __getitem__(self, key):
        """Query the underlying model if it has the key 'key'.
        
        If it does, return a formatted version of the value. 
        If it does not, throw exception 'KeyError'
        """
        # This will raise a KeyError exception if the model
        # does not support the key 'key':
        v = self.model[key]
        # If we made it this far, the model does have key 'key'.
        # Is it a scalar?
        if v is None or isinstance(v, float) or isinstance(v, int):
            # Yes, it is. Format it and return the string.
            str = self.formatter.format(v, self.context + (key,))
            return str
        # It's not a scalar. Create a new instance recursively:
        return ModelView(v, self.formatter, self.context + (key,))
    
    def __getattr__(self, attr):
        """Query the underlying model if it has the attribute 'attr'.
        
        If it does, return a formatted version of the value of the attribute.
        if it does not, throw exception 'AttributeError'.
        """
        # This will raise an AttributeError exception if the model
        # does not support the attribute 'attr':
        v = self.model.__getattribute__(attr)
        # If we made it this far, the model does have attribute 'attr'.
        # Is it a scalar?
        if v is None or isinstance(v, float) or isinstance(v, int) :
            # Yes, it is. Format it and return the string.
            str = self.formatter.format(v, self.context + (attr,))
            return str
        # It's not a scalar. Create a new instance recursively:
        return ModelView(v, self.formatter, self.context + (attr,))
    
class Formatter(object):
    """Formatter to be used to convert a value to a string, given a context.
    
    This is not the most elegant solution as it hardwires in how dates
    and times are treated, but it gets most things right and can be easily
    extended.
    
    """
    def __init__(self, value_format_dict, unit_label_dict, time_format_dict):
        """Initialize an instance of Formatter.
        
        value_label_dict: A dictionary with key of a type, value of a format to be
        used to format that type. Example: {'outTemp' : '%0.1f', 'barometer' : '%0.3f'}
        
        unit_label_dict: A dictionary with the unit label to be used for a type.
        Example: {'outTemp' : '\xb40F', 'barometer': 'inHg'}
        
        time_format_dict: A dictionary with strftime type formats. It should have
        three keys: 'date_format', 'time_format', and 'datetime_format', to be used for
        formatting involving only a date (e.g., '1-May-2007'), only a time
        (e.g., "12:15"), and involving a date and time (e.g., '1-May-2007 12:15'),
        respectively.
        Example: {'date_format' : '%d-%b-%Y', 'time_format' : '%H:%M', 
                  'datetime_format' : '%d-%b-%Y %H:%M'}
        
        """
        self.value_format_dict = value_format_dict
        self.unit_label_dict   = unit_label_dict
        self.time_format_dict  = time_format_dict
        
    def format(self, v, context):
        """Format the value v in the context context.
        
        v: The value to be formatted. If no suitable format can be found, 
        the value itself will be returned.
        
        context: A tuple with the context of the value.
        An example context would be ('year', 'barometer', 'max'). The algorithm
        searches backwards down the context ('max', 'barometer', 'year' in this
        case), looking for something it can format.
        
        returns formatted version of v
        """
        # Walk backwards through the context tuple, looking
        # for something we recognize and can format:
        for _key in context[::-1]:
            # Is it a date?
            if any(_key == type for type  in ('mintime', 'maxtime', 'time', 'dateTime')) :
                # Format as a date/time
                if context[0] == 'day' :
                    _format = 'time_format'
                else :
                    _format = 'datetime_format'
                return time.strftime(self.time_format_dict[_format], time.localtime(v))
            else:
                # Not a date. Is it in one of the other dictionaries?
                if self.value_format_dict.has_key(_key) or self.unit_label_dict.has_key(_key):
                    # It is. Format as a number
                    if v is None :
                        return "N/A"
                    try:
                        val_str   = self.value_format_dict.get(_key, '%f') % v
                    except TypeError:
                        val_str = str(v)
                    label_str = self.unit_label_dict.get(_key, '')
                    return (val_str + label_str).decode('string_escape')
        # Don't know how to format it. Just return it as is
        return v




if __name__ == '__main__' :
    from Cheetah.Template import Template
    class MinMax(object):
        def __init__(self, mn, mx, avg):
            self.min = mn
            self.max = mx
            self.avg = avg
            self.mintime = 1234567890
            self.maxtime = 1235678901
            
    class StatsModel(dict):
        def __init__(self):
            self['day'] = {}
            self['year'] = {}
            self['year']['outTemp'] = MinMax(45.32, 61.3, 48.23431)
            self['year']['barometer'] = MinMax(29.923, 30.8902, 30.23108)
            self['day']['outTemp'] = MinMax(47.1, 51.1, 48.987)
            
    format_dict = {'outTemp' : '%0.1f', 'barometer' : '%0.3f'}
    label_dict  = {'outTemp' : '\xb40F', 'barometer' : ' inHg'}
    date_format_dict = {'date_format' : '%d-%b-%Y', 'time_format' : '%H:%M', 
                        'datetime_format' : '%d-%b-%Y %H:%M'}

    formatter = Formatter(format_dict, label_dict, date_format_dict) 
    stats = StatsModel()
    statsView = ModelView(stats, formatter)
    
#    v1 = statsView['year']
#    v2 = statsView['year']['outTemp']
    v = statsView['year']['barometer'].max
    print "**************", v

    templateDef = '''Max Temperature = $year.outTemp.max\n'''\
                  '''Max Barometer = $year.barometer.max'''
                  
    template=Template(templateDef, searchList = [statsView])

    print template
